#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
File containing definitions for the capability version routes of the Workspaces REST API
"""

import copy
import http
import json
import logging

import requests
from pycapo import CapoConfig
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPPreconditionFailed
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

logger = logging.getLogger(__name__)

from workspaces.capability.enums import CapabilityVersionState


@view_config(route_name="view_all_versions", renderer="json")
def view_all_versions(request: Request):
    """
    Pyramid view that accepts a request to view all versions of a capability request
    URL: capability/request/{capability_request_id}/versions

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.versions:
            return capability_request.versions
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_latest_version", renderer="json")
def view_latest_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])
    if capability_request:
        if capability_request.current_version:
            version_json = capability_request.current_version.__json__()
            version_json["files"] = []

            for file in capability_request.current_version.files:
                version_json["files"].append(file.__json__())

            return Response(json_body=version_json)
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_specific_version", renderer="json")
def view_specific_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            return version
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_internal_notes", renderer="json")
def view_internal_notes(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get notes for a specific version
    URL: capability/request/{capability_request_id}/version/{version_id}/notes

    :param request: GET request
    :return: 200 OK response with JSON-formatted info of version notes
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            return Response(status_int=http.HTTPStatus.OK, json_body={"resp": f"{version.internal_notes}"})
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="update_internal_notes", renderer="json")
def update_internal_notes(request: Request) -> Response:
    """
    Pyramid view that accepts a request to update notes for a specific version
    URL: capability/request/{capability_request_id}/version/{version_id}/notes

    :param request: POST request
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]
    params = request.json_body
    # sanitize input before writing/persisting
    # \\u0000 is an invalid character that is incompatible with postgres json columns
    # from StackOverflow: https://stackoverflow.com/questions/63092267/how-to-handle-api-responsesjson-containing-x00-or-u0000-in-its-data-and-s
    new_notes = json.loads(json.dumps(params["edits"])).replace("\\u0000", "")

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if capability_request:
        if version := capability_request.versions[int(version_id) - 1]:
            version.internal_notes = new_notes
            return Response(
                body=f"Successfully updated internal_notes for capability request #{capability_request_id} v{version_id}."
            )
        else:
            no_versions_msg = f"Capability request with ID {capability_request_id} has no version {version_id}"
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = f"Capability request with ID {capability_request_id} not found."
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="create_capability_version", renderer="json")
def create_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a capability version
    URL: capability/request/{capability_request_id}/version/create

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """
    expected_params = ["parameters"]
    capability_request_id = request.matchdict["capability_request_id"]
    params = request.json_body

    if not all([expected in params for expected in expected_params]):
        # JSON params do not contain all expected params
        params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif not request.capability_info.lookup_capability_request(capability_request_id):
        # CapabilityRequest with given id does not exist; can't create a new version for it
        does_not_exist_msg = (f"Capability request #{capability_request_id} does not exist. Cannot create new version",)
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        new_capability_version = request.capability_info.create_capability_version(
            capability_request_id, parameters=params["parameters"]
        )
        return Response(json_body=new_capability_version.__json__())


@view_config(route_name="submit_capability_version", renderer="json")
def submit_capability_version(request: Request) -> Response:
    """
    Pyramid view that accepts a request to submit a capability version for execution
    URL: capability/request/{capability_request_id}/version/{version_id}/submit

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability request with given id does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_number = request.matchdict["version_number"]

    version = request.capability_info.lookup_version(capability_request_id, version_number)

    if not version_number:
        # Capability request not found
        does_not_exist_msg = (
            f"Capability request #{capability_request_id} v{version_number} does not exist; cannot submit it."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)

    execution = request.capability_launcher.launch_capability(version)
    return Response(json_body=execution.__json__())


@view_config(route_name="cancel_capability_version", renderer="json")
def cancel_capability_version(request: Request) -> Response:
    """
    Accepts a capability request id and version id and cancels it
    URL: capability/request/{capability_request_id}/version/{version_id}/cancel

    :return: 200 OK response
        or 412 response (HTTPPreconditionFailed) if capability execution with the given ids does not exist
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_number = request.matchdict["version_number"]

    version = request.capability_info.lookup_version(capability_request_id, version_number)
    execution = request.capability_info.lookup_execution_by_request_version(capability_request_id, version_number)
    if execution:
        request.capability_canceller.cancel_capability_execution(execution)
        return Response(body=f"Capability request #{capability_request_id} v{version_number} successfully cancelled.")
    elif version:
        if version.state == "Created":
            cap_request = request.capability_info.lookup_capability_request(version.capability_request_id)
            # Set version to Cancelled state
            version.state = CapabilityVersionState.Cancelled.name

            # Seal both version and requests
            version.sealed = True
            cap_request.sealed = True

            # Save
            request.capability_info.save_version(version)
            request.capability_info.save_entity(cap_request)
            return Response(
                body=f"Capability request #{capability_request_id} v{version_number} successfully cancelled."
            )
    else:
        does_not_exist_msg = (
            f"Capability request #{capability_request_id} v{version_number} has not been executed; cannot cancel it."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)


@view_config(route_name="update_file_for_latest", renderer="json")
def add_or_update_file(request: Request) -> Response:
    """
    Pyramid view that accepts a request to edit an existing file to the latest version of a capability request
    URL: capability/request/{capability_request_id}/version/latest/files/{filename}

    :param request: PUT request
    :return: 200 OK response with JSON-formatted info of latest version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["capability_request_id"])

    if capability_request:
        if capability_request.current_version:
            versionFile = request.capability_info.lookup_version_file(
                request.matchdict["capability_request_id"],
                capability_request.current_version.version_number,
                request.matchdict["filename"],
            )
            if versionFile:
                versionFile.content = request.body
                file = request.capability_info.save_edited_version_file(versionFile)
                return file
            else:
                # Add the file if it doesn't exist:
                return request.capability_info.save_version_file(
                    capability_request.current_version, request.matchdict["filename"], request.body
                )
        else:
            no_versions_msg = (
                f"Capability request with ID {request.matchdict['capability_request_id']} has no versions."
            )
            return HTTPPreconditionFailed(no_versions_msg)
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['capability_request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_version_metadata", renderer="json")
def view_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/workflow_metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    capability_request = request.capability_info.lookup_capability_request(capability_request_id)
    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        return version.workflow_metadata
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)


@view_config(route_name="edit_version_metadata", renderer="json")
def edit_version_metadata(request: Request) -> Response:
    """
    Pyramid view that accepts a request to get the workflow metadata of a specific version of a capability request
    URL: capability/request/{capability_request_id}/version/{version_id}/workflow_metadata

    :param request: GET request
    :return: 200 OK response with JSON metadata of specified version of request with given ID
        or 404 response (HTTPNotFound) if capability request with given ID does not exist
        or 412 response (HTTPPreconditionFailed) if capability request with given ID has no versions
    """

    capability_request_id = request.matchdict["capability_request_id"]
    version_id = request.matchdict["version_id"]

    if version := request.capability_info.lookup_version(capability_request_id, version_id):
        # sqlalchemy does not detect in-place mutations of JSON
        # https://docs.sqlalchemy.org/en/14/core/type_basics.html#sqlalchemy.types.JSON
        # to get past this we use a deepcopy of version.workflow_metadata to create a "new" JSON obj
        version.workflow_metadata = copy.deepcopy(version.workflow_metadata)
        if version.workflow_metadata is None:
            version.workflow_metadata = {}
        for key in request.json_body:
            version.workflow_metadata[key] = request.json_body[key]
        request.capability_info.save_version(version)
        return request.json_body
    else:
        no_versions_msg = f"Capability request {capability_request_id} version {version_id} not found."
        return HTTPNotFound(detail=no_versions_msg)
